/*
 * Copyright 2002, Tobias Huerlimann. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#define INCL_DOSERRORS
#define INCL_DOSQUEUES
#define INCL_DOSPROCESS
#include <os2.h>

#include "pmmpgp.h"

handles		my_handles;
options		my_opts;
pgp_commands	pgp_cmds;
pgp_options	pgp_opts;
gpg_call	gpg_cmdline;
char		envvar_lang[] = "LANG=en";

APIRET	pmmpgp_config(void);
APIRET	pmmpgp_setup(void);
APIRET	pmmpgp_exit(void);
void	pmmpgp_print_pgp_cmdline(const int, char**);
APIRET	pmmpgp_build_gpg_args(void);
int	pmmpgp_run_gpg(void);

int main(int argc, char** argv) {
	int	rc = 2;

	if (pmmpgp_setup() != NO_ERROR)
		goto cleanup;
	if (pmmpgp_parse_pgp_cmdline(argc, argv) != NO_ERROR)
		goto cleanup;
	/*
	 * Do not print out the cmdline before pmmpgp_parse_pgp_cmdline()
	 * because pmmpgp_parse_pgp_cmdline() also makes the passphrase
	 * unreadable by replacing it by 'XXXX'.
	 * Otherwise the passphrase would be written to the logfile
	 * in CLEARTEXT!
	 */
	pmmpgp_print_pgp_cmdline(argc, argv);
	if (pmmpgp_build_gpg_args() != NO_ERROR)
		goto cleanup;

	rc = pmmpgp_run_gpg();

cleanup:
	pmmpgp_exit();

	return rc;
}


void APIENTRY pipethread(PVOID param_ptr) {
	APIRET		rc;
	int		i;
	ULONG		ulRead;
	char		line[PMMPGP_MAXLINELEN + 1];
	char		buf[PMMPGP_MAXLINELEN + 1];
	unsigned int	line_pos = 0, buf_pos = 0;

	bzero(&line, sizeof(line));
	bzero(&buf, sizeof(buf));

	while ((rc = DosRead(my_handles.status.read, &buf[buf_pos], sizeof(buf) - buf_pos, &ulRead)) == NO_ERROR) {
		if (ulRead == 0)
			break;
		for (i = 0; i < ulRead; i++, buf_pos++) {
			if (buf[buf_pos] == '\r')
				continue;
			if (buf[buf_pos] == '\n') {
				line[line_pos] = '\0';
				pmmpgp_gpgstatusline(line);
				line_pos = 0;
			} else
				line[line_pos++] = buf[buf_pos];
		}

		DosSleep(10);
	}
	rc = DosClose(my_handles.status.read);

	DosExit(EXIT_THREAD, 0);
}


int	pmmpgp_run_gpg(void) {
	APIRET		rc = NO_ERROR;
	char		load_error[CCHMAXPATH];
	RESULTCODES	resultcodes;
	char		arg[PMMPGP_MAXLINELEN + 2];
	char		*argp = (char*)&arg;
	unsigned int	len;
	TID		pipetid;
	ULONG		ulWritten, ulRead;

	if (my_opts.verbose >= 1)
		fprintf(my_handles.log, "pmmpgp: GPG command line: \"%s %s %s %s %s\"\n",
			gpg_cmdline.executable, gpg_cmdline.opts, gpg_cmdline.rcpts,
			gpg_cmdline.cmds, gpg_cmdline.files);
	fflush(my_handles.log);

	/* Build an argument list suitable for DosExecPgm() */
	bzero(&arg, sizeof(arg));
	strlcpy(arg, gpg_cmdline.executable, sizeof(arg));
	len = strlen(arg);
	argp += len + 1;
	*argp++ = ' ';
	/* "- 3" because of one ' ' and two '\0' chars */
	len = sizeof(arg) - len - 3;
	strlcpy(argp, gpg_cmdline.opts, len);
	strlcat(argp, gpg_cmdline.rcpts, len);
	strlcat(argp, gpg_cmdline.cmds, len);
	strlcat(argp, gpg_cmdline.files, len);

	pipetid = (APIRET)_beginthread(&pipethread, NULL, 32768, NULL);
	if (pipetid == (APIRET)-1) {
		perror("pmmpgp: _beginthread");
		return (APIRET)-1;
	}

	rc = DosWrite(my_handles.pass.write, &pgp_opts.passphrase,
		strlen(pgp_opts.passphrase), &ulWritten);
	if (rc != NO_ERROR) {
		fprintf(my_handles.log, "pmmpgp: DosWrite() failed (rc = %u)\n", rc);
		return -1;
	}
	rc = DosClose(my_handles.pass.write);

	rc = DosExecPgm(load_error, CCHMAXPATH, EXEC_SYNC, (PSZ)&arg,
		(PSZ)&gpg_cmdline.env, &resultcodes, (PSZ)&gpg_cmdline.executable);
	if (rc != NO_ERROR) {
		fprintf(my_handles.log, "pmmpgp: DosExecPgm() failed (rc = %u)\n", rc);
		gpg_cmdline.rc = 254;
		return gpg_cmdline.rc;
	}
	if (my_opts.verbose >= 1)
		fprintf(my_handles.log, "pmmpgp: GnuPG rc = %u\n", resultcodes.codeResult);
	gpg_cmdline.rc = resultcodes.codeResult;

	rc = DosClose(my_handles.status.write);

	rc = DosWaitThread(&pipetid, DCWW_WAIT);
	fflush(my_handles.log);

	if (my_opts.verbose >= 1)
		fprintf(my_handles.log, "pmmpgp: pmmpgp rc = %u\n", gpg_cmdline.rc);
	return gpg_cmdline.rc;
}


ULONG pmmpgp_config(void) {
	APIRET		rc = NO_ERROR;
	char		filename[PMMPGP_MAXPATHLEN + 1] = "";
	FILE*		cfg;
	char		line[PMMPGP_MAXLINELEN + 1];
	char		*option, *value_ptr;
	char		value[PMMPGP_MAXVALUELEN + 1];
	int		value_int;
	int		i, j;

	bzero(&line, sizeof(line));

	/* Set default options */
	my_opts.verbose = 1;
	strlcpy(my_opts.logfile, "pmmpgp.log", sizeof(my_opts.logfile));
	strlcpy(my_opts.gpg_executable, "gpg.exe", sizeof(my_opts.gpg_executable));

	if (getenv("HOME") != NULL)
		strlcpy(filename, getenv("HOME"), sizeof(filename));
	strlcat(filename, "/", sizeof(filename));
	strlcat(filename, PMMPGP_CONFIGFILE, sizeof(filename));
	cfg = fopen(filename, "r");
	if (cfg != NULL) {
		while(fgets(line, sizeof(line), cfg) != NULL) {
			if (line[0] == '#')
				/* Line is a comment */
				continue;
			option = (char*)&line;
			i = 0;
			while ((line[i] != '\0') && (line[i] != ' ') && (line[i] != '\t') && (line[i] != '\r') && (line[i] != '\n'))
				i++;
			if (i == 0)
				/* It's an empty line */
				continue;
			if (line[i] == '\0')
				value_ptr = NULL;
			else {
				line[i++] = '\0';
				value_ptr = (char*)&line[i];
				i = 0;
				j = 0;
				while ((value_ptr[i] != '\0') && (value_ptr[i] != '\r') && (value_ptr[i] != '\n') && (j < PMMPGP_MAXVALUELEN)) {
					switch (value_ptr[i]) {
						case ' ':
						case '\t':
							if (j > 0)
								/* Stop here */
								value_ptr[i] = '\0';
							else
								i++;
							break;
						case '\\':
							i++;
							switch (value_ptr[i++]) {
								case '\"':
									value[j++] = '\"';
									break;
								case '\\':
								default:
									value[j++] = '\\';
									break;
							}
							break;
						case '\"':
							i++;
							while ((value_ptr[i] != '\0') && (value_ptr[i] != '\r') && (value_ptr[i] != '\n') && (value_ptr[i] != '\"') && (j < PMMPGP_MAXVALUELEN))
								if (value_ptr[i] == '\\') {
									i++;
									switch (value_ptr[i++]) {
										case '\"':
											value[j++] = '\"';
											break;
										case '\\':
										default:
											value[j++] = '\\';
											break;
									}
           							} else
									value[j++] = value_ptr[i++];
							/* Skip the terminating double quote */
							if (value_ptr[i] == '\"')
								i++;
							break;
						default:
							value[j++] = value_ptr[i++];
							break;
					}
				}
				value[j] = '\0';
				if (stricmp(option, "verbose") == 0) {
					value_int = atoi(value);
					my_opts.verbose = value_int;
				} else
				if (stricmp(option, "logfile") == 0) {
					strlcpy(my_opts.logfile, value, sizeof(my_opts.logfile));
				} else
				if (stricmp(option, "gpg_exe") == 0) {
					strlcpy(my_opts.gpg_executable, value, sizeof(my_opts.gpg_executable));
				}
			}
		}
		fclose(cfg);
	}

	return rc;
}


ULONG pmmpgp_setup(void) {
	APIRET		rc = NO_ERROR;
	ULONG		action, position;
	time_t		now;
	char		timestamp[30];

	putenv(envvar_lang);
	bzero(&my_handles,	sizeof(my_handles));
	bzero(&my_opts,		sizeof(my_opts));
	bzero(&pgp_opts,	sizeof(pgp_opts));
	bzero(&pgp_cmds,	sizeof(pgp_cmds));
	bzero(&gpg_cmdline,	sizeof(gpg_cmdline));

	rc = pmmpgp_config();
	if (rc != NO_ERROR)
		return rc;

	/* Create a file handle pointing to /dev/null */
	rc = DosOpen("nul", &my_handles.null, &action, 0UL, FILE_NORMAL,
		OPEN_ACTION_FAIL_IF_NEW | OPEN_ACTION_OPEN_IF_EXISTS,
		OPEN_ACCESS_READWRITE | OPEN_SHARE_DENYNONE, NULL);
	if (rc != NO_ERROR)
		return rc;

	my_handles.io.in = stdin;
	my_handles.io.out = stdout;
	my_handles.io.err = stderr;

	/* If the user wants a logfile, open it */
	if (my_opts.logfile[0] != '\0')
		my_handles.log = fopen(my_opts.logfile, "a+");
	else
		my_handles.log = fopen("nul", "a+");
	if (my_handles.log == NULL)
		return (APIRET)-1;

	now = time(NULL);
	strftime(timestamp, 30, "%Y-%m-%d %H:%M:%S", localtime(&now));
	fprintf(my_handles.log, "===== pmmpgp === Start @ %s =================\n", timestamp);

	/* Create the pipe used for passing the passphrase */
	rc = DosCreatePipe(&my_handles.pass.read, &my_handles.pass.write, PMMPGP_PIPESIZE);
	if (rc != NO_ERROR) {
		fprintf(my_handles.log, "pmmpgp: DosCreatePipe() failed (rc = %u)\n", rc);
		return rc;
	}

	/* Create the pipe used to gather status information */
	rc = DosCreatePipe(&my_handles.status.read, &my_handles.status.write, PMMPGP_PIPESIZE);
	if (rc != NO_ERROR) {
		fprintf(my_handles.log, "pmmpgp: DosCreatePipe() failed (rc = %u)\n", rc);
		return rc;
	}

	return rc;
}


ULONG pmmpgp_exit(void) {
	APIRET		rc = NO_ERROR;

	rc = DosClose(my_handles.pass.read);
	rc = DosClose(my_handles.pass.write);

	rc = DosClose(my_handles.status.read);
	rc = DosClose(my_handles.status.write);

	if (my_handles.log != NULL) {
		fprintf(my_handles.log, "\n");
		fclose(my_handles.log);
	}
	rc = DosClose(my_handles.null);

	return rc;
}


void	pmmpgp_print_pgp_cmdline(const int argc, char** argv) {
	int		i;

	if (my_opts.verbose >= 1) {
		fprintf(my_handles.log, "pmmpgp: PGP command line: \"");
		for (i = 0; i < argc; i++)
			fprintf(my_handles.log, ((i + 1 == argc) ? "%s" : "%s "), argv[i]);
		fprintf(my_handles.log, "\"\n");
	}
}


#define ADD_OPT(opt)		strlcat(gpg_cmdline.opts, opt, sizeof(gpg_cmdline.opts))
#define ADD_CMD(cmd)		strlcat(gpg_cmdline.cmds, cmd, sizeof(gpg_cmdline.cmds))
#define ADD_ENVVAL(var, val)	{ \
					snprintf(temp, PMMPGP_MAXVALUELEN, "%s=%s", var, val); \
					len = strlen(temp); \
					strlcat(env, temp, env_avail); \
					env += len + 1; \
					env_avail -= len - 1; \
				}
#define ADD_ENV(var)		value = getenv(var); \
				if (value != NULL) \
					ADD_ENVVAL(var, value);

APIRET	pmmpgp_build_gpg_args(void) {
	int		i;
	char		temp[PMMPGP_MAXVALUELEN + 1];
	char		*value, *env = (char*)&gpg_cmdline.env;
	unsigned int	len, env_avail = sizeof(gpg_cmdline.env) - 1;

	bzero(&temp, sizeof(temp));

	/* Name of the executable */
	strlcpy(gpg_cmdline.executable, my_opts.gpg_executable, sizeof(gpg_cmdline.executable));

	/* Always use the "--quiet" option */
	ADD_OPT("--quiet ");

	/* Options */
	if (pgp_opts.batchmode)
		ADD_OPT("--batch ");
	else
		if (pgp_opts.interactive)
			ADD_OPT("--interactive --no-batch ");
	if (pgp_opts.armor)
		ADD_OPT("--armor ");
	if (pgp_opts.cert_depth != 0) {
		sprintf(temp, "%u ", pgp_opts.cert_depth);
		ADD_OPT("--max-cert-depth ");
		ADD_OPT(temp);
	}
	if (pgp_opts.comment != NULL) {
		ADD_OPT("--comment ");
		ADD_OPT(pgp_opts.comment);
		ADD_OPT(" ");
	}
	if (pgp_opts.charset != NULL)
		if (strcmp(pgp_opts.charset, "noconv") != 0) {    	
			ADD_OPT("--charset ");
			ADD_OPT(pgp_opts.charset);
			ADD_OPT(" ");
		}
	if (pgp_opts.completes_needed != 0) {
		sprintf(temp, "%u ", pgp_opts.completes_needed);
		ADD_OPT("--completes-needed ");
		ADD_OPT(temp);
	}
	if ((!pgp_opts.compress) && (!pgp_opts.armor) && (pgp_cmds.encrypt))
		ADD_OPT("--compress 0 ");
	if (pgp_opts.encrypttoself) {
		/*
		 * FIXME: "--encrypt-to ..."
		 */
	}
	if (pgp_opts.force)
		ADD_OPT("--yes ");
	if (pgp_opts.marginals_needed != 0) {
		sprintf(temp, "%u ", pgp_opts.marginals_needed);
		ADD_OPT("--marginals-needed ");
		ADD_OPT(temp);
	}
	if (pgp_opts.myname != NULL) {
		ADD_OPT("--local-user ");
		ADD_OPT(pgp_opts.myname);
		ADD_OPT(" ");
	}
	if (pgp_opts.pubring != NULL) {
		ADD_OPT("--no-default-keyring --keyring ");
		ADD_OPT(pgp_opts.pubring);
		ADD_OPT(" ");
	}
	if (pgp_opts.secring != NULL) {
		ADD_OPT("--secret-keyring ");
		ADD_OPT(pgp_opts.secring);
		ADD_OPT(" ");
	}
	if (pgp_opts.textmode)
		ADD_OPT("--textmode ");
	if (pgp_opts.verbose > 0)
		for (i = 0; i < pgp_opts.verbose; i++)
			ADD_OPT("--verbose ");
	if (pgp_opts.outputfile[0] != '\0') {
			ADD_OPT("--output ");
			ADD_OPT(pgp_opts.outputfile);
			ADD_OPT(" ");
	} else
		if (pgp_opts.filter)
			ADD_OPT("--output - ");
	if (pgp_opts.userid[0] != '\0') {
			ADD_OPT("--default-key ");
			ADD_OPT(pgp_opts.userid);
			ADD_OPT(" ");
	}
	if (pgp_opts.passphrase[0] != '\0') {
			sprintf(temp, "%u ", my_handles.pass.read);
			ADD_OPT("--passphrase-fd ");
			ADD_OPT(temp);
	}

	/* Add the options to get status information */
	sprintf(temp, "--logger-fd %u --status-fd %u ", fileno(my_handles.log), my_handles.status.write);
	ADD_OPT(temp);

	/* Commands */
	if (pgp_cmds.sign) {
		if (pgp_opts.seperate_signature)
			ADD_CMD("--detach-sign ");
		else
			if ((pgp_opts.clearsig) && (!pgp_cmds.encrypt))
				ADD_CMD("--clearsign ");
			else
				ADD_CMD("--sign ");
	}
	if (pgp_cmds.encrypt)
		ADD_CMD("--encrypt ");
	if (pgp_cmds.decrypt)
		ADD_CMD("--decrypt ");
	if (pgp_cmds.key_add)
		ADD_CMD("--import ");
	if (pgp_cmds.key_gen)
		ADD_CMD("--gen-key ");
	if (pgp_cmds.key_maint)
		ADD_CMD("--check-trustdb ");
	if (pgp_cmds.key_remove)
		ADD_CMD("--delete-key ");
	if (pgp_cmds.key_display)
		ADD_CMD("--list-keys ");
	if (pgp_cmds.key_display_more)
		ADD_CMD("--list-sigs ");
	if (pgp_cmds.key_display_fingerprint)
		ADD_CMD("--fingerprint ");
	if (pgp_cmds.key_extract)
		ADD_CMD("--export ");

	/* Environment variables */
	ADD_ENV("UNIXROOT");
	ADD_ENV("HOME");
	ADD_ENV("GNUPGHOME");
	ADD_ENV("GPG_AGENT_INFO");
	ADD_ENV("http_proxy");
	ADD_ENVVAL("LANG", pgp_opts.language);

	return 0;
}

#undef ADD_ENV
#undef ADD_ENVVAR
#undef ADD_CMD
#undef ADD_OPT

