/*
 * Copyright 2002, Tobias Huerlimann. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#define INCL_DOSERRORS
#define INCL_DOSQUEUES
#define INCL_DOSPROCESS
#include <os2.h>

#include "pmmpgp.h"

#define OT_BOOL			0
#define OT_INT			1
#define OT_STRING		2

#define OPT_ARMOR		0
#define OPT_ARMORLINES		1
#define OPT_BAKRING		2
#define OPT_BATCHMODE		3
#define OPT_CERT_DEPTH		4
#define OPT_CHARSET		5
#define OPT_CLEARSIG		6
#define OPT_COMMENT		7
#define OPT_COMPLETES_NEEDED	8
#define OPT_COMPRESS		9
#define OPT_ENCRYPTTOSELF	10
#define OPT_FORCE		11
#define OPT_INTERACTIVE		12
#define OPT_KEEPBINARY		13
#define OPT_LANGUAGE		14
#define OPT_LEGAL_KLUDGE	15
#define OPT_MAKERANDOM		16
#define OPT_MARGINALS_NEEDED	17
#define OPT_MYNAME		18
#define OPT_NOMANUAL		19
#define OPT_PAGER		20
#define OPT_PUBRING		21
#define OPT_RANDSEED		22
#define OPT_SECRING		23
#define OPT_SHOWPASS		24
#define OPT_TEXTMODE		25
#define OPT_TMP			26
#define OPT_TZFIX		27
#define OPT_VERBOSE		28
#define OPT_NOOPT		29

typedef struct _long_option {
	char	name[PMMPGP_MAXNAMELEN + 1];
	int	opttype;
	char	value[PMMPGP_MAXVALUELEN + 1];
	void*	variable;
} long_option;

long_option	long_opts[OPT_NOOPT] = {
	{"ARMOR",		OT_BOOL,	"", &pgp_opts.armor},
	{"ARMORLINES",		OT_INT,		"", &pgp_opts.armorline},
	{"BAKRING",		OT_STRING,	"", &pgp_opts.bakring},
	{"BATCHMODE",		OT_BOOL,	"", &pgp_opts.batchmode},
	{"CERT_DEPTH",		OT_INT,		"", &pgp_opts.cert_depth},
	{"CHARSET",		OT_STRING,	"", &pgp_opts.charset},
	{"CLEARSIG",		OT_BOOL,	"", &pgp_opts.clearsig},
	{"COMMENT",		OT_STRING,	"", &pgp_opts.comment},
	{"COMPLETES_NEEDED",	OT_BOOL,	"", &pgp_opts.completes_needed},
	{"COMPRESS",		OT_BOOL,	"", &pgp_opts.compress},
	{"ENCRYPTTOSELF",	OT_BOOL,	"", &pgp_opts.encrypttoself},
	{"FORCE",		OT_BOOL,	"", &pgp_opts.force},
	{"INTERACTIVE",		OT_BOOL,	"", &pgp_opts.interactive},
	{"KEEPBINARY",		OT_BOOL,	"", &pgp_opts.keepbinary},
	{"LANGUAGE",		OT_STRING,	"", &pgp_opts.language},
	{"LEGAL_KLUDGE",	OT_INT,		"", &pgp_opts.legal_kludge},
	{"MAKERANDOM",		OT_BOOL,	"", &pgp_opts.makerandom},
	{"MARGINALS_NEEDED",	OT_INT,		"", &pgp_opts.marginals_needed},
	{"MYNAME",		OT_STRING,	"", &pgp_opts.myname},
	{"NOMANUAL",		OT_BOOL,	"", &pgp_opts.nomanual},
	{"PAGER",		OT_STRING,	"", &pgp_opts.pager},
	{"PUBRING",		OT_STRING,	"", &pgp_opts.pubring},
	{"RANDSEED",		OT_STRING,	"", &pgp_opts.randseed},
	{"SECRING",		OT_STRING,	"", &pgp_opts.secring},
	{"SHOWPASS",		OT_BOOL,	"", &pgp_opts.showpass},
	{"TEXTMODE",		OT_BOOL,	"", &pgp_opts.textmode},
	{"TMP",			OT_STRING,	"", &pgp_opts.tmp},
	{"TZFIX",		OT_INT,		"", &pgp_opts.tzfix},
	{"VERBOSE",		OT_INT,		"", &pgp_opts.verbose} };

int interpret_long_opts(void);
int lookup_long_opt(const char*);

int pmmpgp_parse_pgp_cmdline(const int argc, char** argv) {
	int i, j;
	int lo;
	char* current_opt;
	char opt_string[PMMPGP_MAXNAMELEN + 1];
	int opt_string_len;
	int num_pgp_opts;
	int done;
	int file_given = FALSE;

	for (i = 1; i < argc; i++) {
		done = FALSE;
		current_opt = argv[i];
		switch (*current_opt) {
			case '+':
				if (strchr(current_opt, '=') == NULL) {
					strlcpy(opt_string, current_opt + 1, sizeof(opt_string));
					lo = lookup_long_opt(opt_string);
					if (lo != -1) {
						if (long_opts[lo].opttype == OT_BOOL) {
							strlcpy(long_opts[lo].value, "on", PMMPGP_MAXVALUELEN + 1);
						}
					}
				} else {
					opt_string_len = (int)(strchr(current_opt, '=') - current_opt - 1);
					memcpy(opt_string, current_opt + 1, opt_string_len);
					opt_string[opt_string_len] = '\0';
					lo = lookup_long_opt(opt_string);
					if (lo != -1) {
						strlcpy(long_opts[lo].value, current_opt + opt_string_len + 2, PMMPGP_MAXVALUELEN + 1);
					}
				}
				break;
			case '-':
				num_pgp_opts = strlen(current_opt) - 1;
				for (j = 0; j < num_pgp_opts; j++) {
					if (done)
						continue;
					switch (current_opt[j + 1]) {
						case 'a':
							pgp_opts.armor = TRUE;
							break;
						case 'l':
							pgp_opts.verbose = 2;
							break;
						case 'f':
							pgp_opts.filter = TRUE;
							break;
						case 'o':
							if (j == num_pgp_opts - 1) {
								strlcpy(pgp_opts.outputfile, argv[++i], sizeof(pgp_opts.outputfile));
							} else {
								strlcpy(pgp_opts.outputfile, (char*)&current_opt[j + 2], sizeof(pgp_opts.outputfile));
								done = TRUE;
							}
							break;
						case 'z':
							if (j == num_pgp_opts - 1) {
								strlcpy(pgp_opts.passphrase, argv[++i], sizeof(pgp_opts.passphrase));
								/* Make the passphrase string in argv unreadable */
								if (strlen(argv[i]) < 4)
									memset(argv[i], 'X', strlen(argv[i]));
								else
									strlcpy(argv[i], "XXXX", strlen(argv[i]) + 1);
							} else {
								strlcpy(pgp_opts.passphrase, (char*)&current_opt[j + 2], sizeof(pgp_opts.passphrase));
								/* Make the passphrase string in argv unreadable */
								if (strlen((char*)&current_opt[j + 2]) < 4)
									memset((char*)&current_opt[j + 2], 'X', strlen((char*)&current_opt[j + 2]));
								else
									strlcpy((char*)&current_opt[j + 2], "XXXX", strlen((char*)&current_opt[j + 2]) + 1);
								done = TRUE;
							}
							break;
						case 'u':
							if (j == num_pgp_opts - 1) {
								strlcpy(pgp_opts.userid, argv[++i], sizeof(pgp_opts.userid));
							} else {
								strlcpy(pgp_opts.userid, (char*)&current_opt[j + 2], sizeof(pgp_opts.userid));
								done = TRUE;
							}
							break;
						case 'b':
							pgp_opts.seperate_signature = TRUE;
							break;
						case 'm':
							pgp_opts.foryoureyesonly = TRUE;
							break;
						case 'p':
							pgp_opts.preserve_filename = TRUE;
							break;
						case 't':
							pgp_opts.textmode = TRUE;
							break;

						case '?':
						case 'h':
							pgp_cmds.help = TRUE;
							break;
						case 'c':
							pgp_cmds.conventional_encrypt = TRUE;
							break;
						case 'd':
							pgp_cmds.decrypt = TRUE;
							break;
						case 'e':
							pgp_cmds.encrypt = TRUE;
							break;
						case 's':
							pgp_cmds.sign = TRUE;
							break;
						case 'w':
							pgp_cmds.wipe = TRUE;
							break;
						case 'k':
							if (j == num_pgp_opts - 1) {
								pgp_cmds.key_help = TRUE;
							} else {
								switch (current_opt[++j + 1]) {
									case 'a':
										pgp_cmds.key_add = TRUE;
										break;
									case 'c':
										pgp_cmds.key_check = TRUE;
										break;
									case 'd':
										pgp_cmds.key_disable = TRUE;
										break;
									case 'e':
										pgp_cmds.key_edit = TRUE;
										break;
									case 'g':
										pgp_cmds.key_gen = TRUE;
										break;
									case 'm':
										pgp_cmds.key_maint = TRUE;
										break;
									case 'r':
										pgp_cmds.key_remove = TRUE;
										break;
									case 's':
										pgp_cmds.key_sign = TRUE;
										break;
									case 'v':
										if (j == num_pgp_opts - 1) {
											pgp_cmds.key_display = TRUE;
										} else {
											if (current_opt[j + 2] == 'v') {
												pgp_cmds.key_display_more = TRUE;
												j++;
											} else
											if (current_opt[j + 2] == 'c') {
												pgp_cmds.key_display_fingerprint = TRUE;
												j++;
											} else
												pgp_cmds.key_display = TRUE;
										}
										break;
									case 'x':
										pgp_cmds.key_extract = TRUE;
										break;
								}
							}
							break;
					}
				}
				break;
			default:
				if (pgp_cmds.encrypt) {
					if ((file_given) || (pgp_opts.filter)) {
						strlcat(gpg_cmdline.rcpts, "--recipient ", sizeof(gpg_cmdline.rcpts));
						strlcat(gpg_cmdline.rcpts, argv[i], sizeof(gpg_cmdline.rcpts));
						strlcat(gpg_cmdline.rcpts, " ", sizeof(gpg_cmdline.rcpts));
					} else {
						file_given = TRUE;
						strlcat(gpg_cmdline.files, argv[i], sizeof(gpg_cmdline.files));
						strlcat(gpg_cmdline.files, " ", sizeof(gpg_cmdline.files));
					}
				} else {
					if (!pgp_cmds.key_add || (pgp_cmds.key_add && !pgp_opts.filter)) {
						strlcat(gpg_cmdline.files, argv[i], sizeof(gpg_cmdline.files));
						strlcat(gpg_cmdline.files, " ", sizeof(gpg_cmdline.files));
					}
				}
				break;
		}
	}

	interpret_long_opts();

	return;
}


int interpret_long_opts(void) {
	int i;

	for (i = 0; i < OPT_NOOPT; i++) {
		if (*long_opts[i].value != '\0') {
			switch (long_opts[i].opttype) {
				case OT_BOOL:
					if (stricmp(long_opts[i].value, "on") == 0)
						*(char*)long_opts[i].variable = TRUE;
					else
						*(char*)long_opts[i].variable = FALSE;
					break;
				case OT_INT:
					*(int*)long_opts[i].variable = atoi(long_opts[i].value);
					break;
				case OT_STRING:
					*(char**)long_opts[i].variable = (char*)&long_opts[i].value;
					break;
			}
		}
	}

	return 0;
}


int lookup_long_opt(const char* optname) {
	int best_opt = -1;
	int best_result = 0;
	int i, j;
	int str_len, opt_len;

	str_len = strlen(optname);
	for (i = 0; i < OPT_NOOPT; i++) {
		opt_len = strlen(long_opts[i].name);
		if (opt_len < str_len)
			continue;
		for (j = 0; j < str_len; j++) {
			if (toupper(optname[j]) != long_opts[i].name[j])
				break;
		}
		if (j == str_len) {
			best_result = j;
			best_opt = i;
		}		
	}

	return best_opt;
}

